// ignore_for_file: avoid_print

import 'dart:convert';
import 'dart:core';
import 'dart:math' as math;
import 'dart:typed_data';

import 'package:args/args.dart';
import 'package:encrypt/encrypt.dart' as encrypt;

void main(List<String> args) {
  final parser = ArgParser()
    ..addOption('appid', abbr: 'a')
    ..addOption('secret', abbr: 's')
    ..addOption('userid', abbr: 'u');

  final argResults = parser.parse(args);
  final appid = int.parse(argResults['appid']);
  final secret = argResults['secret'];
  final userid = argResults['userid'];
  generateZIMKitToken(appid, secret, userid).then((value) => print(value));
}

Future<String> generateZIMKitToken(
  int appID,
  String appSecret,
  String userID, {
  int effectiveTimeInSeconds = 60 * 60 * 24,
  String payload = '',
}) async {
  if (appID == 0) {
    throw Exception('appID Invalid');
  }
  if (userID == '') {
    throw Exception('userID Invalid');
  }
  if (appSecret.length != 32) {
    throw Exception('appSecret Invalid');
  }
  if (effectiveTimeInSeconds <= 0) {
    throw Exception('effectiveTimeInSeconds Invalid');
  }
  final tokenInfo = TokenInfo04(
    appID: appID,
    userID: userID,
    nonce: math.Random().nextInt(math.pow(2, 31).toInt()),
    ctime: DateTime.now().millisecondsSinceEpoch ~/ 1000,
    expire: 0,
    payload: payload,
  );
  tokenInfo.expire = tokenInfo.ctime + effectiveTimeInSeconds;
  // Convert token information to json
  final tokenJson = tokenInfo.toJson();

  // Randomly generated 16-byte string, used as AES encryption vector,
  // before the ciphertext for Base64 encoding to generate the final token
  final ivStr = ZIMKitStringUtils.createRandomString(16);
  final iv = encrypt.IV.fromUtf8(ivStr);

  final key = encrypt.Key.fromUtf8(appSecret);
  final encrypter = encrypt.Encrypter(encrypt.AES(key, mode: encrypt.AESMode.cbc));
  final encrypted = encrypter.encrypt(tokenJson, iv: iv);

  final bytes1 = HexUtils.createUInt8ListFromInt(tokenInfo.expire);
  final bytes2 = Uint8List.fromList([0, 16]);
  final bytes3 = Uint8List.fromList(utf8.encode(ivStr));
  final bytes4 = Uint8List.fromList([0, encrypted.bytes.length]);
  final bytes5 = encrypted.bytes;

  final bytes = Uint8List(4) + bytes1 + bytes2 + bytes3 + bytes4 + bytes5;
  final ret = '04${base64.encode(bytes)}';
  return ret;
}

class TokenInfo04 {
  TokenInfo04({
    required this.appID,
    required this.userID,
    required this.ctime,
    required this.expire,
    required this.nonce,
    required this.payload,
  });

  int appID;
  String userID;
  int nonce;
  int ctime;
  int expire;
  String payload;

  String toJson() {
    return '{"app_id":$appID,"user_id":"$userID","nonce":$nonce,"ctime":$ctime,"expire":$expire,"payload":"$payload"}';
  }
}

class ZIMKitStringUtils {
  static final _random = math.Random();
  static const _defaultPool = 'ModuleSymbhasOwnPr-0123456789ABCDEFGHNRVfgctiUvz_KqYTJkLxpZXIjQW';

  static String createRandomString(int size, {String pool = _defaultPool}) {
    final len = pool.length;
    var id = '';
    var i = size;
    while (0 < i--) {
      id += pool[_random.nextInt(len)];
    }
    return id;
  }

  static Future<String> createRandomStringAsync(int size, {String pool = _defaultPool}) async {
    return createRandomString(size, pool: pool);
  }
}

extension ZIMKitStringUtilsExtension on String {
  String get urlEncode {
    return replaceAll('#', '%23').replaceAll('&', '%26');
  }
}

class HexUtils {
  /// Creates a `Uint8List` by a hex string.
  static Uint8List createUInt8ListFromHexString(String hex) {
    final result = Uint8List(hex.length ~/ 2);
    for (var i = 0; i < hex.length; i += 2) {
      final num = hex.substring(i, i + 2);
      final byte = int.parse(num, radix: 16);
      result[i ~/ 2] = byte;
    }

    return result;
  }

  /// Returns a hex string by a `Uint8List`.
  static String formatBytesAsHexString(Uint8List bytes) {
    final result = StringBuffer();
    for (var i = 0; i < bytes.lengthInBytes; i++) {
      final part = bytes[i];
      result.write('${part < 16 ? '0' : ''}${part.toRadixString(16)}');
    }

    return result.toString();
  }

  static Uint8List createUInt8ListFromInt(int hex) {
    return createUInt8ListFromHexString(hex.toRadixString(16));
  }
}

class ZIMKitTokenInfo04 {
  ZIMKitTokenInfo04({
    required this.appID,
    required this.userID,
    required this.ctime,
    required this.expire,
    required this.nonce,
    required this.payload,
  });

  int appID;
  String userID;
  int nonce;
  int ctime;
  int expire;
  String payload;

  String toJson() {
    return '{"app_id":$appID,"user_id":"$userID","nonce":$nonce,"ctime":$ctime,"expire":$expire,"payload":"$payload"}';
  }
}
